# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
import authenticator_pb2 as authenticator__pb2
import grpc


class AuthenticatorStub(object):
    """Missing associated documentation comment in .proto file"""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.GetPGMD5Hash = channel.unary_unary(
                '/approzium.authenticator.protos.Authenticator/GetPGMD5Hash',
                request_serializer=authenticator__pb2.PGMD5HashRequest.SerializeToString,
                response_deserializer=authenticator__pb2.PGMD5Response.FromString,
                )
        self.GetPGSHA256Hash = channel.unary_unary(
                '/approzium.authenticator.protos.Authenticator/GetPGSHA256Hash',
                request_serializer=authenticator__pb2.PGSHA256HashRequest.SerializeToString,
                response_deserializer=authenticator__pb2.PGSHA256Response.FromString,
                )
        self.GetMYSQLSHA1Hash = channel.unary_unary(
                '/approzium.authenticator.protos.Authenticator/GetMYSQLSHA1Hash',
                request_serializer=authenticator__pb2.MYSQLSHA1HashRequest.SerializeToString,
                response_deserializer=authenticator__pb2.MYSQLSHA1Response.FromString,
                )


class AuthenticatorServicer(object):
    """Missing associated documentation comment in .proto file"""

    def GetPGMD5Hash(self, request, context):
        """Missing associated documentation comment in .proto file"""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetPGSHA256Hash(self, request, context):
        """Missing associated documentation comment in .proto file"""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetMYSQLSHA1Hash(self, request, context):
        """Missing associated documentation comment in .proto file"""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_AuthenticatorServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'GetPGMD5Hash': grpc.unary_unary_rpc_method_handler(
                    servicer.GetPGMD5Hash,
                    request_deserializer=authenticator__pb2.PGMD5HashRequest.FromString,
                    response_serializer=authenticator__pb2.PGMD5Response.SerializeToString,
            ),
            'GetPGSHA256Hash': grpc.unary_unary_rpc_method_handler(
                    servicer.GetPGSHA256Hash,
                    request_deserializer=authenticator__pb2.PGSHA256HashRequest.FromString,
                    response_serializer=authenticator__pb2.PGSHA256Response.SerializeToString,
            ),
            'GetMYSQLSHA1Hash': grpc.unary_unary_rpc_method_handler(
                    servicer.GetMYSQLSHA1Hash,
                    request_deserializer=authenticator__pb2.MYSQLSHA1HashRequest.FromString,
                    response_serializer=authenticator__pb2.MYSQLSHA1Response.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'approzium.authenticator.protos.Authenticator', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class Authenticator(object):
    """Missing associated documentation comment in .proto file"""

    @staticmethod
    def GetPGMD5Hash(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/approzium.authenticator.protos.Authenticator/GetPGMD5Hash',
            authenticator__pb2.PGMD5HashRequest.SerializeToString,
            authenticator__pb2.PGMD5Response.FromString,
            options, channel_credentials,
            call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetPGSHA256Hash(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/approzium.authenticator.protos.Authenticator/GetPGSHA256Hash',
            authenticator__pb2.PGSHA256HashRequest.SerializeToString,
            authenticator__pb2.PGSHA256Response.FromString,
            options, channel_credentials,
            call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetMYSQLSHA1Hash(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/approzium.authenticator.protos.Authenticator/GetMYSQLSHA1Hash',
            authenticator__pb2.MYSQLSHA1HashRequest.SerializeToString,
            authenticator__pb2.MYSQLSHA1Response.FromString,
            options, channel_credentials,
            call_credentials, compression, wait_for_ready, timeout, metadata)



import logging
import select

import psycopg2

import approzium

from .._postgres import PGAuthClient
from ._psycopg2_ctypes import (
    ensure_compatible_ssl,
    libpq_PQstatus,
    read_msg,
    set_connection_sync,
    set_debug,
    write_msg,
)

logger = logging.getLogger(__name__)


def wait(pgconn):
    while True:
        state = pgconn.poll()
        if state == psycopg2.extensions.POLL_OK:
            break
        elif state == psycopg2.extensions.POLL_WRITE:
            select.select([], [pgconn.fileno()], [])
        elif state == psycopg2.extensions.POLL_READ:
            select.select([pgconn.fileno()], [], [])
        else:
            raise psycopg2.OperationalError("poll() returned %s" % state)


def construct_approzium_conn(is_sync, authenticator):
    class ApproziumConn(psycopg2.extensions.connection):
        CONNECTION_AWAITING_RESPONSE = 4

        def __init__(self, *args, **kwargs):
            logger.debug("ApproziumConn __init__")
            kwargs.pop("async", None)
            kwargs.pop("async_", None)
            super().__init__(*args, **kwargs, async_=1)
            if self.dsn is None:
                # connection is uninitalized due to an error
                return
            if logger.getEffectiveLevel() <= logging.DEBUG:
                set_debug(self)
            self.authenticator = authenticator
            dbhost = self.get_dsn_parameters()["host"]
            dbport = self.get_dsn_parameters()["port"]
            dbuser = self.get_dsn_parameters()["user"]
            self._pgauthclient = PGAuthClient(
                lambda: read_msg(self),
                lambda msg: write_msg(self, msg),
                authenticator,
                dbhost,
                dbport,
                dbuser,
            )
            self._checked_ssl = False
            if is_sync:
                wait(self)
                set_connection_sync(self)
                self.autocommit = False

        def poll(self):
            status = libpq_PQstatus(self.pgconn_ptr)
            if self.info.ssl_in_use and not self._checked_ssl:
                ensure_compatible_ssl(self)
                logging.debug("checked ssl")
                self._checked_ssl = True
            if (
                status == self.CONNECTION_AWAITING_RESPONSE
                and not self._pgauthclient.done
            ):
                next(self._pgauthclient)
                return psycopg2.extensions.POLL_WRITE
            else:
                logging.debug("normal poll")
                return super().poll()

    return ApproziumConn


def connect(dsn=None, cursor_factory=None, authenticator=None, **kwargs):
    """Creates a Psycopg2 connection through Approzium authentication. Takes
    the same arguments as ``psycopg2.connect``, in addition to the
    authenticator argument.

    :param authenticator: AuthClient instance to be used for authentication. If
        not provided, the default AuthClient, if set, is used.
    :type authenticator: approzium.AuthClient, optional
    :raises: TypeError, if no AuthClient is given and no default one is set.
    :rtype: ``psycopg2.Connection``

    Example:

    .. code-block:: python

        >>> import approzium
        >>> from approzium.psycopg2 import connect
        >>> auth = approzium.AuthClient("myauthenticator.com:6001", disable_tls=True)
        >>> con = connect("host=DB.com dbname=mydb", authenticator=auth)  # no password!
        >>> # use the connection just like any other Psycopg2 connection

    .. warning::
        Currently, only `psycopg2` with dynamically linked `libpq` is
        supported. Thus, `psycopg2-binary` is not supported.
    """
    is_sync = True
    if kwargs.get("async", False):
        is_sync = False
    if kwargs.get("async_", False):
        is_sync = False
    if authenticator is None:
        authenticator = approzium.default_auth_client
    if authenticator is None:
        raise TypeError("Auth client not specified and not default auth client is set")
    factory = construct_approzium_conn(is_sync, authenticator)
    return psycopg2.connect(dsn, factory, cursor_factory, **kwargs)



from contextlib import contextmanager

import mysql.connector
from mysql.connector import MySQLConnection

import approzium

from ..._mysql import get_auth_resp_msg


class ApproziumMySQLConnection(MySQLConnection):
    def _do_auth(self, *args, **kwargs):
        if self._password.__class__.__name__ == "AuthClient":

            def _auth_response(
                client_flags,
                username,
                password,
                database,
                auth_plugin,
                auth_data,
                ssl_enabled,
            ):
                authenticator = password
                is_secure_connection = (
                    client_flags
                    & mysql.connector.constants.ClientFlag.SECURE_CONNECTION
                )
                auth_response = get_auth_resp_msg(
                    is_secure_connection,
                    authenticator,
                    host,
                    str(port),
                    username,
                    auth_plugin,
                    auth_data,
                )
                return auth_response

            host = self.server_host
            port = self.server_port
            self._protocol._auth_response = _auth_response
            self.authenticator = self._password

        return super(ApproziumMySQLConnection, self)._do_auth(*args, **kwargs)


@contextmanager
def _patch_MySQLConnection(include_pooling=False):
    mysql.connector.MySQLConnection = ApproziumMySQLConnection
    if include_pooling:
        mysql.connector.pooling.MySQLConnection = ApproziumMySQLConnection
    try:
        yield
    finally:
        mysql.connector.MySQLConnection = MySQLConnection
        if include_pooling:
            mysql.connector.pooling.MySQLConnection = MySQLConnection


def _parse_kwargs(kwargs):
    authenticator = kwargs.pop("authenticator", None)
    if authenticator is None:
        authenticator = approzium.default_auth_client
    if authenticator is None:
        raise TypeError("Auth client not specified and no default auth client is set")
    kwargs["password"] = authenticator
    use_pure = kwargs.get("use_pure", False)
    if not use_pure:
        msg = "MySQL C-Extension based connection is not currently supported.\
Pass use_pure=True"
        raise NotImplementedError(msg)
    return kwargs


def connect(*args, authenticator=None, **kwargs):
    """Creates a MySQL connector connection through Approzium authentication. Takes
    the same arguments as ``mysql.connector.connect``, in addition to the
    authenticator argument.

    :param authenticator: AuthClient instance to be used for authentication. If
        not provided, the default AuthClient, if set, is used.
    :type authenticator: approzium.AuthClient, optional
    :raises: TypeError, if no AuthClient is given and no default one is set.
    :rtype: ``mysql.connector.MySQLConnection``

    Example:

    .. code-block:: python

        >>> import approzium
        >>> from approzium.mysql.connector import connect
        >>> auth = approzium.AuthClient("myauthenticator.com:6001", disable_tls=True)
        >>> con = connect(user="bob", host="host.com", authenticator=auth, \
        ...     use_pure=True)
        >>> # use the connection just like any other MySQL connector connection

    .. warning::
        Currently, only the pure Python MySQL connector implementation is
        supported. Therefore, you have to pass in ``use_pure=True``, otherwise,
        an exception is raised.

    .. note:
        Currently, only secure password authentication using
        `mysql_native_password` is supported. For more details, read MySQL's
        `documentation`_

    .. _documentation:
        https://dev.mysql.com/doc/internals/en/secure-password-authentication.html
    """
    kwargs = _parse_kwargs({"authenticator": authenticator, **kwargs})
    with _patch_MySQLConnection():
        conn = mysql.connector.connect(*args, **kwargs)
    return conn



from mysql.connector.pooling import MySQLConnectionPool

from ._connect import _parse_kwargs, _patch_MySQLConnection


class MySQLConnectionPool(MySQLConnectionPool):
    def set_config(self, **kwargs):
        kwargs = _parse_kwargs(kwargs)
        super(MySQLConnectionPool, self).set_config(**kwargs)

    def add_connection(self, cnx=None):
        with _patch_MySQLConnection(include_pooling=True):
            super().add_connection(cnx)



import logging

from ._authenticator import AuthClient, TLSConfig

default_auth_client = None
"""
Set this variable to an instance of :class:`~approzium.AuthClient` to set it as
the default auth client to be used for connections.
"""

logger = logging.getLogger(__name__)
log_format = "[%(filename)s:%(lineno)s - %(funcName)10s() ] %(message)s"
formatter = logging.Formatter(log_format)
ch = logging.StreamHandler()
ch.setFormatter(formatter)
logger.addHandler(ch)


__all__ = ["default_auth_client", "AuthClient", "TLSConfig"]



import logging
import os
import select
import struct
import subprocess
from ctypes import (
    CDLL,
    c_char_p,
    c_int,
    c_void_p,
    cdll,
    create_string_buffer,
    memmove,
    string_at,
)
from ctypes.util import find_library
from os import path
from sys import getsizeof

logger = logging.getLogger(__name__)

libpq = cdll.LoadLibrary(find_library("pq"))


# setup ctypes functions
# necessary to avoid segfaults when using multiple Python threads
libpq_PQstatus = libpq.PQstatus
libpq_PQstatus.argtypes = [c_void_p]
libpq_PQstatus.restype = c_int

libpq_PQsslInUse = libpq.PQsslInUse
libpq_PQsslInUse.argtypes = [c_void_p]
libpq_PQsslInUse.restype = c_int

libpq_PQgetssl = libpq.PQgetssl
libpq_PQgetssl.argtypes = [c_void_p]
libpq_PQgetssl.restype = c_void_p

libpq_PQsetnonblocking = libpq.PQsetnonblocking
libpq_PQsetnonblocking.argtypes = [c_void_p, c_int]
libpq_PQsetnonblocking.restype = c_int


def stdout(command):
    return subprocess.run(command, capture_output=True).stdout.decode("utf-8")


def ssl_supported():
    out = stdout(["pg_config", "--configure"])
    return "--with-openssl" in out


def possible_library_files(name):
    return [
        "lib%s.dylib" % name,
        "%s.dylib" % name,
        "%s.framework/%s" % (name, name),
        "lib%s.so" % name,
    ]


def setup_ssl():
    sslpath = ""
    # try to find OpenSSL path in `pg_config`'s LDFLAGS
    out = stdout(["pg_config", "--ldflags"])
    for lib in out.split(" "):
        if "openssl" in lib:
            ssldir = lib.split("-L")[-1]
            # directory path is found, so search for actual file
            for filename in possible_library_files("ssl"):
                possible_sslpath = path.join(ssldir, filename)
                if path.exists(possible_sslpath):
                    sslpath = possible_sslpath
                    break
    # if none is found, use the SSL library that the system's dynamic linker finds
    if not sslpath:
        sslpath = find_library("ssl")
    global libssl
    global libssl_SSL_read
    global libssl_SSL_write
    libssl = cdll.LoadLibrary(sslpath)
    libssl_SSL_read = libssl.SSL_read
    libssl_SSL_read.argtypes = [c_void_p, c_char_p, c_int]
    libssl_SSL_read.restype = c_int

    libssl_SSL_write = libssl.SSL_write
    libssl_SSL_write.argtypes = [c_void_p, c_char_p, c_int]
    libssl_SSL_write.restype = c_int


def set_connection_sync(pgconn):
    mem = bytearray(string_at(id(pgconn), getsizeof(pgconn)))
    sizeofint = struct.calcsize("@i")
    sizeoflong = struct.calcsize("@l")

    def addressofint(number, mem=mem):
        int_bytes = struct.pack("@i", number)
        return mem.find(int_bytes)

    def intataddress(address):
        return struct.unpack("@i", mem[address : address + sizeofint])[0]

    def ensure(check):
        if not check:
            raise Exception("Could not set connection to sync. Unidentified struct")

    # as a check, we check server and protocol version numbers, which succeed
    # the async value in the psycopg connection struct
    server_version_addr = addressofint(pgconn.server_version)
    # check that there is only one match for that value
    ensure(
        addressofint(pgconn.server_version, mem[server_version_addr + sizeofint :])
        == -1
    )
    protocol_address = server_version_addr - sizeofint
    protocol_version = intataddress(protocol_address)
    ensure(protocol_version == pgconn.protocol_version)
    async_address = protocol_address - sizeoflong
    async_value = struct.unpack("@l", mem[async_address:protocol_address])[0]
    ensure(async_value == pgconn.async_)
    new_async_value = struct.pack("@l", 0)
    memmove(id(pgconn) + async_address, new_async_value, sizeoflong)
    ensure(pgconn.async_ == 0)
    error = libpq_PQsetnonblocking(pgconn.pgconn_ptr, 0)
    ensure(error == 0)


def read_msg(pgconn):
    def read_bytes(n):
        if pgconn.info.ssl_in_use:
            buffer = bytearray(n)
            c_buffer = create_string_buffer(bytes(buffer), n)
            ssl_obj = libpq_PQgetssl(pgconn.pgconn_ptr)
            nread = -1
            while nread == -1:
                nread = libssl_SSL_read(ssl_obj, c_buffer, n)

            msg = bytes(c_buffer.raw[:nread])
            return msg
        else:
            fd = pgconn.fileno()
            return os.read(fd, n)

    select.select([pgconn.fileno()], [], [])
    msg_type = read_bytes(1)
    msg_size_b = read_bytes(4)
    msg_size = struct.unpack("!i", msg_size_b)[0]
    msg_content = read_bytes(msg_size - 4)
    logger.debug(f"read: {msg_type} {msg_content}")
    return msg_type + msg_size_b + msg_content


def write_msg(pgconn, msg):
    select.select([], [pgconn.fileno()], [])
    if libpq_PQsslInUse(pgconn.pgconn_ptr):
        ssl_obj = libpq_PQgetssl(pgconn.pgconn_ptr)
        c_buffer = create_string_buffer(msg, len(msg))
        n = libssl_SSL_write(ssl_obj, c_buffer, len(msg))
        if n != len(msg):
            raise ValueError("could not send response")
    else:
        os.write(pgconn.fileno(), msg)
    logger.debug(f"sent: {msg}")


def set_debug(conn):
    libc = CDLL(find_library("c"))
    stdout = c_void_p.in_dll(libc, "stdout")
    libpq.PQtrace(conn.pgconn_ptr, stdout)


def ensure_compatible_ssl(conn):
    if conn.info.ssl_attribute("library") != "OpenSSL":
        raise Exception("Unsupported SSL library")


if ssl_supported():
    setup_ssl()



import asyncio

import asyncpg
from asyncpg.connection import Connection

from ._asyncpg_connect import connect, new__connect_addr  # approzium's connect method


class _ApproziumPool(asyncpg.pool.Pool):
    async def _get_new_connection(self):
        if self._working_addr is None:
            # First connection attempt on this pool.
            # 1) use our connect method instead of asyncpg's
            con = await connect(
                *self._connect_args, loop=self._loop, **self._connect_kwargs
            )

            self._working_addr = con._addr
            self._working_config = con._config
            self._working_params = con._params
            # 2) after first connection, store class instance because it
            # contains authenticator
            self._connection_class = con.__class__

        else:
            # We've connected before and have a resolved address,
            # and parsed options and config.

            # 3) use our own _connect_addr instead of asyncpg's
            con = await new__connect_addr(
                loop=self._loop,
                addr=self._working_addr,
                timeout=self._working_params.connect_timeout,
                config=self._working_config,
                params=self._working_params,
                connection_class=self._connection_class,
            )

        if self._init is not None:
            try:
                await self._init(con)
            except (Exception, asyncio.CancelledError) as ex:
                # If a user-defined `init` function fails, we don't
                # know if the connection is safe for re-use, hence
                # we close it.  A new connection will be created
                # when `acquire` is called again.
                try:
                    # Use `close()` to close the connection gracefully.
                    # An exception in `init` isn't necessarily caused
                    # by an IO or a protocol error.  close() will
                    # do the necessary cleanup via _release_on_close().
                    await con.close()
                finally:
                    raise ex

        return con


def create_pool(
    dsn=None,
    *,
    min_size=10,
    max_size=10,
    max_queries=50000,
    max_inactive_connection_lifetime=300.0,
    setup=None,
    init=None,
    loop=None,
    authenticator=None,
    **connect_kwargs,
):
    """Create an Asyncpg connection pool through Approzium authentication.
    Takes same arguments as ``asyncpg.create_pool`` in addition to the
    `authenticator` argument

    :return: An instance of :class:`~approzium.asyncpg.pool._ApproziumPool`.

    Example:

    .. code-block:: python

        >>> import approzium
        >>> from approzium.asyncpg import create_pool
        >>> auth = approzium.AuthClient("myauthenticator.com:6001", disable_tls=True)
        >>> pool = await create_pool(user='postgres', authenticator=auth)
        >>> con = await pool.acquire()
        >>> try:
        ...     await con.fetch('SELECT 1')
        ... finally:
        ...     await pool.release(con)
    """
    return _ApproziumPool(
        dsn,
        connection_class=Connection,
        min_size=min_size,
        max_size=max_size,
        max_queries=max_queries,
        loop=loop,
        setup=setup,
        init=init,
        max_inactive_connection_lifetime=max_inactive_connection_lifetime,
        authenticator=authenticator,
        **connect_kwargs,
    )



import psycopg2.pool

from . import connect


class _AbstractConnectionPool(psycopg2.pool.AbstractConnectionPool):
    def _connect(self, key=None):
        # originally, this line uses psycopg2.connect. we change it to use our
        # connect method instead
        conn = connect(*self._args, **self._kwargs)
        if key is not None:
            self._used[key] = conn
            self._rused[id(conn)] = key
        else:
            self._pool.append(conn)
        return conn


class SimpleConnectionPool(_AbstractConnectionPool, psycopg2.pool.SimpleConnectionPool):
    """A Psycopg2 pool that can't be shared across different threads.

    Example:

    .. code-block:: python

        >>> import approzium
        >>> from approzium.psycopg2.pool import SimpleConnectionPool
        >>> auth = approzium.AuthClient("myauthenticator.com:6001", disable_tls=True)
        >>> pool = SimpleConnectionPool("host=DB.com dbname=mydb", authenticator=auth)
        >>> conn = poll.getconn()
    """

    pass


class ThreadedConnectionPool(
    _AbstractConnectionPool, psycopg2.pool.ThreadedConnectionPool
):
    """A Psycopg2 pool that works with the threading module"""

    pass



# needed to be able to import protos code
import json
import sys
from datetime import datetime, timedelta
from itertools import count
from pathlib import Path

import grpc

from . import _iam, _mysql, _postgres
from ._iam import (
    assume_role,
    get_local_arn,
    obtain_claimed_arn,
    obtain_credentials,
    obtain_signed_get_caller_identity,
)

sys.path.append(str(Path(__file__).parent / "_protos"))  # isort:skip
import authenticator_pb2  # noqa: E402 isort:skip
import authenticator_pb2_grpc  # noqa: E402 isort:skip


class AuthClient(object):
    """This class represents a connection to an Approzium authenticator
    service. Instances of this class can be used as arguments to database
    drivers connect method to use for authentication.

    :param server_address: address (host:port) at which an authenticator
        service is listening.
    :type server_address: str

    :param disable_tls: defaults to False. When False, https is used and a
        client_cert and client_key proving the client's identity must be
        provided. When True, http is used and no other TLS options must be
        set.
    :type disable_tls: bool, optional

    :param tls_config: the TLS config to use for encrypted communication.
    :type tls_config: TLSConfig, optional

    :param iam_role: if an IAM role Amazon resource number (ARN) is provided,
        it will be assumed and its identity will be used for authentication.
        Otherwise, the default ``boto3`` session will be used as the identity.
    :type iam_role: str, optional
    """

    def __init__(
        self, server_address, disable_tls=False, tls_config=None, iam_role=None,
    ):
        self.server_address = server_address

        if not disable_tls:
            if tls_config is None:
                raise ValueError("if tls is not disabled, tls config must be provided")
            if tls_config.client_cert is None or tls_config.client_key is None:
                raise ValueError(
                    "if tls is not disabled, "
                    "client_cert and client_key must be provided"
                )
            self.tls_config = TLSConfig(
                trusted_certs=tls_config.trusted_certs,
                client_cert=tls_config.client_cert,
                client_key=tls_config.client_key,
            )

        self.disable_tls = disable_tls
        self.authenticated = False
        self._counter = count(1)
        self.n_conns = 0

        # Parse the claimed ARN once because it'll never change.
        # Parse the signed_gci at startup, and then we'll update
        # it every 5 minutes because it expires every 15.
        if iam_role is None:
            claimed_arn = get_local_arn()
            signed_gci = obtain_signed_get_caller_identity(None)
        else:
            response = assume_role(iam_role)
            credentials = obtain_credentials(response)
            claimed_arn = obtain_claimed_arn(response)
            signed_gci = obtain_signed_get_caller_identity(credentials)

        self.claimed_arn = claimed_arn
        self.signed_gci = signed_gci
        self.signed_gci_last_updated = datetime.utcnow()

    @property
    def attribution_info(self):
        """Provides a dictionary containing information about the current state
        of the AuthClient. Useful for logging.

        :rtype: dict

        **Return Structure**:
            - *authenticator_address* (*str*): address of authenticator service used
            - *iam_arn* (*str*): IAM Amazon resource number (ARN) used as identity
            - *authenticated* (*bool*): whether the AuthClient was verified by the
              authenticator service.
            - *num_connections* (*int*): number of connections made through this
              AuthClient
        """
        info = {}
        info["authenticator_address"] = self.server_address
        info["iam_arn"] = self.claimed_arn
        info["authenticated"] = self.authenticated
        info["num_connections"] = self.n_conns
        info.update(_iam.attribution_info())
        return info

    @property
    def attribution_info_json(self):
        """Provides the same attribution info returned by
        :func:`~AuthClient.attribution_info` as a JSON format string

        :rtype: str
        """
        info = self.attribution_info
        return json.dumps(info)

    def _execute_request(self, request, getmethodname, dbhost, dbport, dbuser):
        # The presigned GetCallerIdentity call expires every 15 minutes.
        self._update_gci_if_needed()

        if self.disable_tls:
            channel = grpc.insecure_channel(self.server_address)
        else:
            credentials = grpc.ssl_channel_credentials(
                root_certificates=_read_file(self.tls_config.trusted_certs),
                certificate_chain=_read_file(self.tls_config.client_cert),
                private_key=_read_file(self.tls_config.client_key),
            )
            channel = grpc.secure_channel(self.server_address, credentials)

        stub = authenticator_pb2_grpc.AuthenticatorStub(channel)
        # authentication info
        aws_identity = authenticator_pb2.AWSIdentity(
            signed_get_caller_identity=self.signed_gci,
            claimed_iam_arn=self.claimed_arn,
        )
        password_request = authenticator_pb2.PasswordRequest(
            aws=aws_identity,
            client_language=authenticator_pb2.PYTHON,
            dbhost=dbhost,
            dbport=dbport,
            dbuser=dbuser,
        )
        request.pwd_request.CopyFrom(password_request)
        response = getattr(stub, getmethodname)(request)
        # if no exception is raised, request was successful
        self.authenticated = True
        self.n_conns = next(self._counter)
        return response

    def _get_pg2_hash(self, dbhost, dbport, dbuser, auth_type, auth_info):
        if auth_type == _postgres.AUTH_REQ_MD5:
            salt = auth_info
            if len(salt) != 4:
                raise Exception("salt not right size")
            request = authenticator_pb2.PGMD5HashRequest(salt=salt,)
            response = self._execute_request(
                request, "GetPGMD5Hash", dbhost, dbport, dbuser
            )
            return response.hash
        elif auth_type == _postgres.AUTH_REQ_SASL:
            auth = auth_info
            auth._generate_auth_msg()
            request = authenticator_pb2.PGSHA256HashRequest(
                salt=auth.password_salt,
                iterations=auth.password_iterations,
                authentication_msg=auth.authorization_message,
            )
            response = self._execute_request(
                request, "GetPGSHA256Hash", dbhost, dbport, dbuser
            )
            client_final = auth.create_client_final_message(response.cproof)
            auth.server_signature = response.sproof
            return client_final, auth

    def _get_mysql_hash(self, dbhost, dbport, dbuser, auth_type, auth_info):
        if auth_type == _mysql.MYSQLNativePassword:
            salt = auth_info
            if len(salt) != 20:
                raise Exception("salt not right size")
            request = authenticator_pb2.MYSQLSHA1HashRequest(salt=salt,)
            response = self._execute_request(
                request, "GetMYSQLSHA1Hash", dbhost, dbport, dbuser
            )
            return response.hash
        else:
            raise Exception("Unexpected authentication method")

    # The presigned GetCallerIdentity string expires every 15 minuts, so refresh it
    # after 5 minutes just to be safe.
    def _update_gci_if_needed(self):
        if datetime.utcnow() - self.signed_gci_last_updated < timedelta(minutes=5):
            return
        if self.iam_role is None:
            self.signed_gci = obtain_signed_get_caller_identity(None)
            self.signed_gci_last_updated = datetime.utcnow()
        else:
            response = assume_role(self.iam_role)
            credentials = obtain_credentials(response)
            self.signed_gci = obtain_signed_get_caller_identity(credentials)
            self.signed_gci_last_updated = datetime.utcnow()


class TLSConfig(object):
    """This class represents the TLS config to be used while communicating
    with Approzium. Its fields are further described here:
    https://grpc.github.io/grpc/python/grpc.html#create-client-credentials

    :param trusted_certs: the path to the root certificate(s) that must have
        issued the identity certificate used by Approzium's authentication
        server.
    :type trusted_certs: str, optional

    :param client_cert: this client's certificate, used for proving its
        identity, and used by the caller to encrypt communication with
        its public key
    :type client_cert: str, optional

    :param client_key: this client's key, used for decrypting incoming
        communication that was encrypted by callers using the client_cert's
        public key
    :type client_key: str, optional
    """

    def __init__(self, trusted_certs=None, client_cert=None, client_key=None):
        self.trusted_certs = trusted_certs
        self.client_cert = client_cert
        self.client_key = client_key


def _read_file(path_to_file):
    with open(path_to_file, "rb") as f:
        return f.read()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['approzium',
 'approzium._mysql',
 'approzium._postgres',
 'approzium.asyncpg',
 'approzium.mysql',
 'approzium.mysql.connector',
 'approzium.opentelemetry',
 'approzium.psycopg2',
 'approzium.pymysql']

package_data = \
{'': ['*'], 'approzium': ['_protos/*']}

install_requires = \
['boto3>=1.14.10,<2.0.0',
 'ec2-metadata>=2.2.0,<3.0.0',
 'grpcio-tools>=1.30.0,<2.0.0',
 'grpcio>=1.30.0,<2.0.0']

extras_require = \
{'sqllibs': ['asyncpg>=0.20.1,<0.22.0',
             'mysql-connector-python>=8.0.20,<9.0.0',
             'psycopg2>=2.8.5,<3.0.0',
             'pymysql>=0.9.3,<0.10.0'],
 'tracing': ['opentelemetry-instrumentation-psycopg2>=0.23b2,<0.24',
             'opentelemetry-exporter-jaeger>=1.4.1,<2.0.0']}

setup_kwargs = {
    'name': 'approzium',
    'version': '0.2.4',
    'description': 'Approzium SDK provides Approzium database authentiation for Python',
    'long_description': None,
    'author': 'Cyral',
    'author_email': 'security@cyral.org',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/cyralinc/approzium',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



import struct

MYSQLNativePassword = "mysql_native_password"


def get_auth_resp(authenticator, dbhost, dbport, dbuser, auth_plugin, auth_data):
    plugin_auth_response = authenticator._get_mysql_hash(
        dbhost, dbport, dbuser, auth_plugin, auth_data
    )
    return plugin_auth_response


def get_auth_resp_msg(is_secure_connection, *args, **kwargs):
    plugin_auth_response = get_auth_resp(*args, **kwargs)
    if is_secure_connection:
        resplen = len(plugin_auth_response)
        auth_response = struct.pack("<B", resplen) + plugin_auth_response
    else:
        auth_response = plugin_auth_response + b"\x00"
    return auth_response
